import numpy as np
import math

dim = 2


def get_function_pacman(str_function, alpha=None):
    if str_function == 'singular':
        assert alpha is not None, 'Provide alpha value for str_function = singular!'

        u = lambda x: np.linalg.norm(x, axis=1, keepdims=True) ** (1 / alpha) \
                      * np.sin(my_arctan(x[:, [0]], x[:, [1]]) / alpha) + 1
        f = lambda x: np.zeros((x.shape[0], 1))
    elif str_function == 'smooth':
        u = lambda x: np.linalg.norm(x, axis=1, keepdims=True) ** 2 + 1
        f = lambda x: -4 * np.ones((x.shape[0], 1))

    return u, f

def my_arctan(x1, x2):
    # Returns angle in the interval [0, 2pi]

    phi = np.arctan2(x2, x1)
    phi += (phi < 0) * 2 * math.pi

    return phi


def sample_domain_pacman(n1, n2, alpha):

    assert 0 <= alpha <= 2, 'alpha must satisfy 0 <= alpha * pi <= 2pi'

    # Create interior
    X1 = 2*np.random.rand(n1, dim) - 1
    X1 = X1[np.linalg.norm(X1, axis=1) < 1]
    X1 = X1[my_arctan(X1[:, 0], X1[:, 1]) < math.pi * alpha, :]      # this is not 100% correct due to points with y=0

    # Create boundary
    array_linspace = np.linspace(0, 1, n2).reshape(-1, 1)
    X2_Gamma1 = np.concatenate((array_linspace, np.zeros_like(array_linspace)), axis=1)
    X2_Gamma2 = np.concatenate((np.cos(array_linspace * alpha * math.pi), np.sin(array_linspace  * alpha * math.pi)), axis=1)
    X2_Gamma3 = np.concatenate((np.cos(alpha * math.pi) * array_linspace,
                                np.sin(alpha * math.pi) * array_linspace), axis=1)

    X2 = np.concatenate((X2_Gamma1, X2_Gamma2, X2_Gamma3), axis=0)


    # Create a meshgrid using numpy.meshgrid
    X1_grid0_, X1_grid1_ = np.meshgrid(np.linspace(-1, 1, 1001), np.linspace(-1, 1, 1001))
    X1_grid = np.concatenate((X1_grid0_.reshape(-1, 1), X1_grid1_.reshape(-1, 1)), axis=1)
    X1_grid = X1_grid[np.linalg.norm(X1_grid, axis=1) < 1, :]                           # shrink to circle
    X1_grid = X1_grid[my_arctan(X1_grid[:, 0], X1_grid[:, 1]) < alpha * math.pi, :]     # shrink to angle

    return X1, X2, X1_grid